-- NiHao by Recluse@Norgannon
--
-- 
--				[[ ***** LOCALS ***** ]]
local itemPrice = AceLibrary("ItemPrice-1.1")
local tablet = AceLibrary("Tablet-2.0")
local L = AceLibrary("AceLocale-2.2"):new("NiHao")
local _G = getfenv(0)
local t = {}
local eventMap = {
	CHAT_MSG_CHANNEL = true,
	CHAT_MSG_SAY = true,
	CHAT_MSG_GUILD = true,
	CHAT_MSG_WHISPER = true,
	CHAT_MSG_WHISPER_INFORM = true,
	CHAT_MSG_YELL = true,
	CHAT_MSG_PARTY = true,
	CHAT_MSG_OFFICER = true,
	CHAT_MSG_RAID = true,
	CHAT_MSG_RAID_LEADER = true,
	CHAT_MSG_RAID_WARNING = true,
	CHAT_MSG_BATTLEGROUND = true,
	CHAT_MSG_BATTLEGROUND_LEADER = true,
	CHAT_MSG_SYSTEM = true,
	CHAT_MSG_DND = true,
	CHAT_MSG_AFK = true,
	CHAT_MSG_LOOT = true,
}
local NULL_INFO = {r = 1.0, g = 1.0, b = 1.0, id = 0 }
local function TableCopy(t, lookup_table)
	local copy = {}
	if type(t) ~= "table" then return t end
	for i,v in pairs(t) do
		if type(v) ~= "table" then
			copy[i] = v
		else
			lookup_table = lookup_table or {}
			lookup_table[t] = copy
			if lookup_table[v] then
				copy[i] = lookup_table[v] -- we already copied this table. reuse the copy.
			else
				copy[i] = TableCopy(v,lookup_table) -- not yet copied. copy it.
			end
		end
	end
	return copy
end
--[[									===TableMigrate===
	Notes:	This function accepts two tables.  It ensures that the old table and the new table
			are identical in layout.
			1. Any keys which exist in both tables have their data preserved in the new table.
			2. Any keys which no longer exist in the new table are removed from the	old table.
				A. Data which existed in those keys will be lost.
			3. Any keys which do not exist in the old table but do in the new table are added
				A. Newly added keys are given a default value set by the new table.
	Reference:
		curT/newT = current and new table.
		curK/newK = current and new key.
		curV/newV = current and new value.
	Example Usage:
		myTable = { a = "foo", b = "bar" }
		myDesiredTable = { a = "red", c = "green" }
		TableMigrate(myTable, myDesiredTable)
		result --> myTable = { a = "foo", c = "green" }
		]]
local function TableMigrate(curT, newT)
	--[[First, we need to add keys to the current table which need to be added.
		If a new key is added, we assign the default value provided by the new table passed in.
		After checking, determine if we are looking at a table, and if so, recurse into it.]]
	for newK, newV in pairs(newT) do
		if curT[newK] == nil then
			curT[newK] = newT[newK]
		end
		if (type(newT[newK]) == "table") then
			TableMigrate(curT[newK], newT[newK])
		end
	end
	--[[Second, we need to remove keys and their values which do not exist in the new table.
		After checking, determine if we are looking at a table, and if so, recurse into it.]]
	for curK, curV in pairs(curT) do
		if newT[curK] == nil then
			curT[curK] = nil
		end
		if (type(curT[curK]) == "table") then
			TableMigrate(curT[curK], newT[curK])
		end
	end
end
local function safestr(s) return s or "" end
--
--
--				[[ ***** NIHAO SETTINGS ***** ]]
NiHao = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0", "AceDB-2.0", "AceEvent-2.0", "AceHook-2.1", "AceConsole-2.0")
NiHao.name = "NiHao"
NiHao.revision = tonumber(string.sub("$Revision: 62261 $", 12, -3))
NiHao.version = "r"..NiHao.revision
NiHao.date = string.sub("$Date: 2008-02-20 13:44:02 -0500 (Wed, 20 Feb 2008) $", 8, 17)
NiHao.hideWithoutStandby = true
NiHao.hasIcon = true
NiHao:RegisterDB("NiHaoDB")
NiHao.session = {	-- Session data storage
	startTime = nil,	-- Time when session began
	pauseTime = nil,	-- Time when session becomes paused
	paused = false,	-- Whether or not the session is paused or active
	flashCounter = 0,	-- Controls the state of the flashing "PAUSED" message
	cashInitial = 0,	-- The player's money when session begins
	cashEarned = 0,		-- The player's earnings of gold, silver, or copper during the session
	itemsTotal = 0,		-- The player's cash value of all items looted during the session
	itemsEarned = {},	-- Populated as the player picks up items
}
NiHao.eventID = {
	current = 0,
	lastProcessed = 0,
}
NiHao.eventData = {}
NiHao.CurrentMsg = nil
NiHao.SplitMessageOut = {
		MESSAGE = "",
		TYPE = "",
		TARGET = "",
		CHANNEL = "",
		LANGUAGE = "",
}
NiHao.SplitMessageOrg = {}
NiHao.INFO = {r = 1.0, g = 1.0, b = 1.0, id = 0 }
NiHao.SplitMessageSrc = {
		PRE = "",

		cC = "",
				CHANNELNUM = "",
				CC = "",
				CHANNEL = "",
				zZ = "",
						ZONE = "",
				Zz = "",
		Cc = "",

		TYPEPREFIX = "",

		fF = "",
				FLAG = "",
		Ff = "",

		pP = "",
				lL = "",	-- link start
				PLAYERLINK= "",
				PLAYERLINKDATA = "",
				LL = "", --	link text start

				PREPLAYER = "",
				PLAYER = "",

				sS = "",
						SERVER = "",
				Ss = "",
				POSTPLAYER = "",

				Ll = "",	-- link text end
		Pp = "",

		TYPEPOSTFIX = "",

		mM = "",
				gG = "",
				LANGUAGE = "",
				Gg = "",
				MESSAGE = "",
		Mm = "",

		POST = "",
}
NiHao.Separators = {
		ZONE = "zz",
		SERVER = "ss",
		MESSAGE = "mm",
		LANGUAGE = "gg",
		PLAYERLINK = "ll",
		FLAG = "ff",
		CHANNEL = "cc",
		PLAYER = "pp",
}
NiHao.SplitMessageIdx = {
		"PRE",
		"cC",
				"CHANNELNUM",
				"CC",
				"CHANNEL",
		"Cc",
		"TYPEPREFIX",
		"fF",
				"FLAG",
		"Ff",
		"pP",
				"lL",
						"PLAYERLINK",
						"PLAYERLINKDATA",
				"LL",
				"PREPLAYER",
				"PLAYER",
				"sS",
						"SERVER",
				"Ss",
				"POSTPLAYER",
				"Ll",
		"Pp",
		"TYPEPOSTFIX",
		"mM",
				"gG",
				"LANGUAGE",
				"Gg",
				"MESSAGE",
		"Mm",
		"POST",
}
NiHao.SplitMessage = {}
NiHao.lootStrings = {
	mobitem = "(You receive loot:).*",
	--mobcash = {
		--prefix	= "You loot (%d+)",	-- If found, we're looking at a cash loot string
		--gold	= "(%d+) Gold",		-- Return amount of gold picked up (if any)
		--silver	= "(%d+) Silver",	-- Return amount of silver picked up (if any)
		--copper	= "(%d+) Copper",	-- Return amount of copper picked up (if any)
	--},
	--questitem = "Received item: [item]",
	--questcash = "Received XX Silver",
}
NiHao.help = {
	color = { 0.5, 0.5, 1.0, nil },
	text = {
		"NiHao Help",
		"   |cFF00FF00\/nihao menu|r - Opens the NiHao FuBar menu. Note: Not all slash commands will show on this menu.",
		" ",
		"   |cFF00FF00\/nihao set [item link] coppervalue|r - Set the value of the item linked, expressed in copper.",
		"   |cFF00FF00\/nihao unset [item link]|r - Remove custom value setting (goes back to vendor values).",
		"   |cFF00FF00\/nihao vendor [item link]|r - Forces the item to use vendor buyback value. Useful when AucAdvanced reports silly values.",
		" ",
	},
}
NiHao.options = {
	type = 'group',
	args = {
		session = {
			type = 'group',
			order = 5,
			name = "Session Settings",
			desc = "Settings",
			args = {
				resetSession = {
					type = 'execute',
					name = "Reset Session",
					desc = "Clears data gathered",
					order = 5,
					func = function() NiHao:ResetSession(true, false) end,
				},
				saveSession = {
					type = 'text',
					name = "Save Session",
					desc = "Saves the current session with the name given",
					usage = "Type the name you would like to save this session as and press the Enter key. Please do not include spaces in the name.",
					order = 6,
					get = function() return nil end,
					set = function(v) NiHao:SaveSession(v) end
				},
				pauseSession = {
					type = 'toggle',
					name = "Pause/Resume Session",
					desc = "Pauses Session (Click again to resume)",
					order = 10,
					get = function() return NiHao.session.paused end,
					set = function(v)
						if NiHao.session.paused then
							NiHao:ResumeSession()
						else
							NiHao:PauseSession(true)
						end
					end,
				},
				startPaused = {
					type = 'toggle',
					name = "Start Paused",
					desc = "If enabled, session will start/restart in a paused state",
					order = 15,
					get = function() return NiHao.db.profile.startPaused end,
					set = function(v) NiHao.db.profile.startPaused = not(NiHao.db.profile.startPaused) end,
				},
				stickySession = {
					type = 'toggle',
					name = "Sticky",
					desc = "If enabled, session data will be carried over into the next session, requiring you to manually reset the session if desired.",
					order = 20,
					get = function() return NiHao.db.profile.stickySession end,
					set = function(v) NiHao.db.profile.stickySession = not(NiHao.db.profile.stickySession) end,		
				},
				savedSessions = {
					type = 'group',
					order = 25,
					name = "Saved Sessions",
					desc = "All sessions you have saved.",
					args = {}
				},
			}
		},
		display = {
			type = 'group',
			order = 10,
			name = "Display Settings",
			desc = "Settings",
			args = {
				captionType = {
					type = 'execute',
					name = "Caption: Total/Per Hour",
					desc = "Toggle FuBar caption to show session total or session per hour.",
					order = 5,
					func = function() NiHao.db.profile.showPerHour = not(NiHao.db.profile.showPerHour) end,
				},
				captionLength = {
					type = 'toggle',
					name = "Short Caption",
					desc = "Remove 'NiHao:' from FuBar caption?",
					order = 10,
					get = function() return NiHao.db.profile.captionLength end,
					set = function() NiHao.db.profile.captionLength = not(NiHao.db.profile.captionLength) end,
				},
				topTen = {
					type = 'toggle',
					name = "Top Loot",
					desc = "Show/Hide the Top Loot listing on the tooltip.",
					order = 15,
					get = function() return NiHao.db.profile.showTopTen end,
					set = function() NiHao.db.profile.showTopTen = not(NiHao.db.profile.showTopTen) end,
				},
				topLootQty = {
					type = 'range',
					name = "Top Loot Qty",
					desc = "How many items to show on the Top Loot list.",
					max = 10,
					min = 1,
					step = 1,
					isPercent = false,
					order = 20,
					get = function() return NiHao.db.profile.topLootQty end,
					set = function(v) NiHao.db.profile.topLootQty = tonumber(v) end,
					hidden = function() return not(NiHao.db.profile.showTopTen) end,
				},
				splitPerHour = {
					type = 'toggle',
					name = "Split per hour",
					desc = "Shows earnings per hour breakdown for item/cash loots.",
					order = 25,
					get = function() return NiHao.db.profile.splitPerHour end,
					set = function() NiHao.db.profile.splitPerHour = not(NiHao.db.profile.splitPerHour) end,
				},
			}
		},
		useAuctioneer = {
			type = 'toggle',
			name = "Use AucAdvanced",
			desc = "Uses Auctioneer Advanced market value prices.  Only applies to items of Common(white) or better quality.  If you have set a custom price for an item, the Auctioneer price will NOT be used. Built for AuctioneerAdvanced version 5.0.PRE.2294 (BillyGoat)",
			order = 15,
			get = function() return NiHao.db.profile.useAuctioneer end,
			set = function() NiHao.db.profile.useAuctioneer = not(NiHao.db.profile.useAuctioneer) end,
			hidden = function() return not(_G.AucAdvanced) end, -- Should hide if user does not have Auctioneer Advanced running.
		},
	},
}
NiHao.OnMenuRequest = NiHao.options
NiHao.itemsIndexed = {}
--
--
--				[[ ***** EVENT HANDLING ***** ]]
function NiHao:OnInitialize()
	self:RegisterAllEvents("OnEvent")
	self:Hook("ChatFrame_MessageEventHandler", true)
	
	-- Call the chat frame hook function so we can start listening to channels.
	self:HookChatFrames()
	
	-- Default profile options (if they are not set, this is what they'll get set to
	self:RegisterDefaults('profile', {
		splitPerHour = false,
		showPerHour = false,
		showTopTen = true,
		startPaused = false,
		captionLength = false,
		useAuctioneer = false,
		itemData = {},
		stickySession = false,
		sessionData = {},
		savedSessions = {},
		topLootQty = 1,
	})
	
	-- Compatibility for no FuBar, though it shouldn't happen XD
	if not FuBar then
		self.OnMenuRequest.args.hide.guiName = L["HideIcon"]
		self.OnMenuRequest.args.hide.desc = L["HideIcon"]
	end
	
	self:ResetSession(false, true)
	self:UpdateSessionMenu()
end
function NiHao:OnEnable()
	self:RegisterChatCommand( { "/NiHao", "/nihao" }, self.SlashHandler, string.upper(self.name) )
	SlashCmdList["NIHAO"] = NiHao_SlashHandler
		SLASH_NIHAO1 = "/nihao"

	self:ResetSession(true, true)
	self:ScheduleRepeatingEvent("NiHaoUpdater", self.UpdateText, 1, self)
	self:UpdateText()
end
function NiHao:OnTooltipUpdate()
	local currentTime = time()
	if (self.session.paused) then
		currentTime = self.session.pauseTime
	end
	
	local timeElapsed = currentTime - self.session.startTime
	local earnings = self:ConvertCash(tonumber(self.session.cashEarned) + tonumber(self.session.itemsTotal), "TEXT")
	local percent = 0
	local itemearningsperhour, cashearningsperhour = nil
	local cashearnings = self:ConvertCash(tonumber(self.session.cashEarned), "TEXT")
	if tonumber(self.session.cashEarned) > 0 then
		percent = math.ceil((tonumber(self.session.cashEarned)/(tonumber(self.session.cashEarned) + tonumber(self.session.itemsTotal)))*100)
	end
	cashearnings = cashearnings.."("..percent.."%)"
	percent = 0
	local itemearnings = self:ConvertCash(tonumber(self.session.itemsTotal), "TEXT")
	if tonumber(self.session.itemsTotal) > 0 then
		percent = math.ceil((tonumber(self.session.itemsTotal)/(tonumber(self.session.cashEarned) + tonumber(self.session.itemsTotal)))*100)
	end
	itemearnings = itemearnings.."("..percent.."%)"
	local earningsperhour = self:ConvertCash(math.ceil(((tonumber(self.session.cashEarned) + tonumber(self.session.itemsTotal))/timeElapsed)* 3600), "TEXT")
	if (NiHao.db.profile.splitPerHour) then
		itemearningsperhour = self:ConvertCash(math.ceil((tonumber(self.session.itemsTotal)/timeElapsed)* 3600), "TEXT")
		cashearningsperhour = self:ConvertCash(math.ceil((tonumber(self.session.cashEarned)/timeElapsed)* 3600), "TEXT")
	end
	local readableTime = self:ConvertTime(timeElapsed)
	local cat = tablet:AddCategory(
		'id', "earnings",
		'columns', 3,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1,
		'child_text3R', 1,
		'child_text3G', 1,
		'child_text3B', 1
	)
	if (self.session.paused) then
		if (self.session.flashCounter == 0) then
			cat:AddLine(
				'text', "|cFFFF9999-PAUSED-|r",
				'text2', "",
				'text3', ""
			)
		else			
			cat:AddLine(
				'text', "      ",
				'text2', "",
				'text3', ""
			)
		end
		cat:AddLine(
			'text', "",
			'text2', "",
			'text3', ""
		)
	end
	cat:AddLine(
		'text', "|cFF9999FFEarnings|r",
		'text2', "",
		'text3', ""
	)
	cat:AddLine(
		'text', "Total",
		'text2', "",
		'text3', earnings
	)
	if NiHao.db.profile.splitPerHour then
		cat:AddLine(
			'text', "Total Per Hour",
			'text2', "",
			'text3', earningsperhour
		)
		cat:AddLine(
			'text', "-Cash Per Hour",
			'text2', "",
			'text3', cashearningsperhour
		)
		cat:AddLine(
			'text', "-Items Per Hour",
			'text2', "",
			'text3', itemearningsperhour
		)
	else
		cat:AddLine(
			'text', "Per Hour",
			'text2', "",
			'text3', earningsperhour
		)
	end
	cat:AddLine(
		'text', "",
		'text2', "",
		'text3', ""
	)
	cat:AddLine(
		'text', "",
		'text2', "",
		'text3', ""
	)
	cat:AddLine(
		'text', "Time Elapsed",
		'text2', "",
		'text3', readableTime
	)
	cat:AddLine(
		'text', "",
		'text2', "",
		'text3', ""
	)
	cat:AddLine(
		'text', "",
		'text2', "",
		'text3', ""
	)
	cat:AddLine(
		'text', "Cash Total",
		'text2', "",
		'text3', cashearnings
	)
	cat:AddLine(
		'text', "Item Total",
		'text2', "",
		'text3', itemearnings
	)
	if NiHao.db.profile.showTopTen then
		cat = tablet:AddCategory(
			'id', "itemlist",
			'columns', 3,
			'child_textR', 1,
			'child_textG', 1,
			'child_textB', 0,
			'child_text2R', 1,
			'child_text2G', 1,
			'child_text2B', 1,
			'child_text3R', 0,
			'child_text3G', 1,
			'child_text3B', 0
		)
		topLootCaption = "|cFF9999FFTop "..tostring(self.db.profile.topLootQty).." Items|r"
		cat:AddLine(
			'text', topLootCaption,
			'text2', "",
			'text3', ""
		)
		local itemCount = 0
		for k,v in pairs(self.session.itemsEarned) do
			itemCount = itemCount + 1
		end
		if (itemCount == 0) then
			cat:AddLine(
				'text', "|cFFFF9999No items looted yet.|r",
				'text2', "",
				'text3', ""
			)
		end
		-- Add list of looted items
		local itemName, itemQty
		topTenCounter = 0
		for k,v in pairs(self.itemsIndexed) do
			if topTenCounter < tonumber(self.db.profile.topLootQty) then -- Restrict to only the top 'x' items as defined by user.
				itemName = self:GetItemData(v, nil, nil, "NAME")
				itemRarity = self:GetItemData(v, nil, nil, "RARITY")
				itemName = ITEM_QUALITY_COLORS[tonumber(itemRarity)].hex..itemName.."|r" -- Colorizes item text by rarity for tooltip
				itemQty = "x"..self.session.itemsEarned[v].quantity
				itemVal = self:ConvertCash(self.session.itemsEarned[v].valueTotal, "TEXT")
				cat:AddLine(
					'text', itemName,
					'text2', itemQty,
					'text3', itemVal
				)
				topTenCounter = topTenCounter + 1
			end
		end
	end
end
function NiHao:OnClick(button)
	-- Toggle display method
	self.db.profile.showPerHour = not self.db.profile.showPerHour
	self:UpdateText()
end
function NiHao:OnTextUpdate()
	local currentTime = time()
	
	-- Optional, as specified by user
	local labelPrefix = "NiHao: "
	if (NiHao.db.profile.captionLength) then
		labelPrefix = ""
	end
	
	-- Override current time with time paused, if NiHao is paused
	if (self.session.paused) then
		currentTime = self.session.pauseTime
	end
	
	-- Determine what information should be displayed
	if self.db.profile.showPerHour == false then
		-- Compile total information
		captionValue = tonumber(self.session.cashEarned) + tonumber(self.session.itemsTotal)
		thelabel = labelPrefix..self:ConvertCash(captionValue, "TEXT")
	else
		-- Compile per hour information
		local timeElapsed = currentTime - self.session.startTime
		captionValue = math.ceil(((tonumber(self.session.cashEarned) + tonumber(self.session.itemsTotal))/timeElapsed)* 3600)
		thelabel = labelPrefix..self:ConvertCash(captionValue, "TEXT").."/hr"
	end
	
	-- Alternate data with paused message if paused
	if (self.session.paused) then
		if (self.session.flashCounter == 1) then
			self.session.flashCounter = 0
			thelabel = labelPrefix.."|cFFFF9999-PAUSED-|r"
		else			
			self.session.flashCounter = self.session.flashCounter + 1
		end
	end
	
	-- Update everything
	self:SetText(thelabel)
	self:SaveSession() -- A check will be done inside of the function
	self:UpdateTooltip() -- Update the tooltip data
	-- Keep updating
	self:ScheduleRepeatingEvent("NiHaoUpdater", self.UpdateText, 1, self)
end
function NiHao:OnDisable()
	self:UnregisterAllEvents()
end
function NiHao:OnEvent(...)
	local event = AceLibrary("AceEvent-2.0").currentEvent
	if type(event) == "string" and strsub(event, 1, 12) == "PLAYER_MONEY" then
		self:UpdateCash()
	end
	if type(event) == "string" and event ~= "CHAT_MSG_ADDON" and strsub(event, 1, 8) == "CHAT_MSG" then
		NiHao.eventData.type = event
		-- Assign a 'serial number' to this event
		NiHao.eventID.current = NiHao.eventID.current + 1
		NiHao.eventData.id = NiHao.eventID.current
		NiHao.eventData.wasChatMsg = true
	end
end
function NiHao_SlashHandler(cmd)
	if not(cmd) then -- in case user didn't type anything after /nihao, this prevents nil value errors
		cmd = ""
	end
	local item, value = nil
	-- Check for set
	_, _, item, value = string.find(cmd, "^set (.+) (%d+)")		-- /nihao set [itemlink] coppervalue
	if item and value then
		value = tonumber(value)
		if value < 1 then value = 0 end -- Prevent user from putting negative amounts.
		local setItemID = NiHao:GetItemData(nil, nil, item, "ID")
		local setItemQty = 0
		if not(NiHao.session.itemsEarned[setItemID]) then -- User never looted this yet
			setItemQty = 0
		elseif not(NiHao.session.itemsEarned[setItemID].quantity) then -- Item in loot list but no quantity data
			NiHao.session.itemsEarned[setItemID].quantity = 0
			setItemQty = tonumber(NiHao.session.itemsEarned[setItemID].quantity)
		else -- Item in loot list with a quantity
			setItemQty = tonumber(NiHao.session.itemsEarned[setItemID].quantity)
		end
		NiHao.db.profile.itemData[setItemID] = {} -- Clear custom data about this item or create a new table for it if it does not exist yet
		NiHao.db.profile.itemData[setItemID].value = value -- Drop in our value
		if (NiHao.session.itemsEarned[setItemID]) then -- If the item exists in the loot list then reassign the new total value of the items looted
			NiHao.session.itemsEarned[setItemID].valueTotal = value * setItemQty
		end
		NiHao:RecalculateItems()
		NiHao:Print("Set the value of each "..NiHao:GetItemData(setItemID, nil, nil, "NAME").." to "..NiHao:ConvertCash(value, "TEXT")..".") -- Inform user of success
		return
	end
	item, value = nil
	-- Check for vendor set
	_, _, item = string.find(cmd, "^vendor (.+)")				-- /nihao vendor [itemlink]
	if item then
		NiHao_SlashHandler("set "..item.." "..NiHao:GetItemData(NiHao:GetItemData(nil, nil, item, "ID"), nil, nil, "VENDOR"))
		return
	end
	item = nil
	-- Check for unset
	_, _, item = string.find(cmd, "^unset (.+)")				-- /nihao unset [itemlink]
	if item then
		if not(NiHao.db.profile.itemData[NiHao:GetItemData(nil, nil, item, "ID")]) then
			NiHao:Print("Cannot remove custom value of "..NiHao:GetItemData(NiHao:GetItemData(nil, nil, item, "ID"), nil, nil, "NAME").." - No custom value is currently set.")
		else
			NiHao.db.profile.itemData[NiHao:GetItemData(nil, nil, item, "ID")] = nil
			if NiHao.session.itemsEarned[NiHao:GetItemData(nil, nil, item, "ID")] then -- user has looted this item
				NiHao.session.itemsEarned[NiHao:GetItemData(nil, nil, item, "ID")].valueTotal = NiHao.session.itemsEarned[NiHao:GetItemData(nil, nil, item, "ID")].valueEach * NiHao.session.itemsEarned[NiHao:GetItemData(nil, nil, item, "ID")].quantity
			end
			NiHao:RecalculateItems()
			NiHao:Print("Removed custom value of "..NiHao:GetItemData(NiHao:GetItemData(nil, nil, item, "ID"), nil, nil, "NAME")..".")
		end
		return
	end
	item = nil
	-- Check for menu
	_, _, menuCommand = string.find(cmd, "^(menu)")				-- /nihao menu
	if menuCommand then
		NiHao:OpenMenu()
		return
	end
	menuCommand = nil
	-- If still here, options bad, print help
	for _, line in ipairs(NiHao.help.text) do
		NiHao:Print(line);
	end
end
--
--
--				[[ ***** CHAT FRAME FUNCTIONS ***** ]]
function NiHao:HookChatFrames()
	-- Loop uses NUM_CHAT_WINDOWS rather than a static number in case Blizzard adds more windows in the future.
	for i=1,NUM_CHAT_WINDOWS do
		-- Create a reference to the chat window based on our loop integer
		local cf = getglobal("ChatFrame"..i);
		-- Hook AddMessage() for this chat window
		self:Hook(cf, "AddMessage", true)
	end
end
function NiHao:ChatFrame_MessageEventHandler(event)
	local event = event
	local message, info
	local process = eventMap[event]
	local eventResult
	message, info = self:SplitChatMessage(this, event)
	if not info then
		self.eventData.infoNil = true
		return self.hooks["ChatFrame_MessageEventHandler"](event)
	else
		if (not(self.eventID.current == self.eventID.lastProcessed)) and (message.CHATTYPE == "LOOT") then
			-- Call to process the text to see if our string was found
			self:ProcessText(message.MESSAGE)
		end
		eventResult = self.hooks["ChatFrame_MessageEventHandler"](event)
	end
	NiHao.eventID.lastProcessed = NiHao.eventID.current
	return eventResult
end
function NiHao:AddMessage(frame, text, r, g, b, id)
	local s = NiHao.SplitMessage
	if s.OUTPUT == nil and s.CAPTUREOUTPUT then
		self.INFO.r, self.INFO.g, self.INFO.b, self.INFO.id = r, g, b, id
		s.OUTPUT = text
				s.INFO = NiHao.INFO
		else
				NiHao.hooks[frame].AddMessage(frame, text, r, g, b, id)
		end
end
--
--
--				[[ ***** ITEMS ***** ]]
function NiHao:UpdateItems(lootText)
	if not(self.session.paused) then
		-- Make sure we have loot text, if not, bail
		local isLootText = nil
		_, _, isLootText = string.find(lootText, "^(You receive loot:).*")
		self.eventData.isLootText = isLootText
		if not(isLootText) then return end
		-- Get information about the item
		local lootLink = self:GetItemLink(lootText)
		local lootItemID = self:GetItemData(nil, nil, lootText, "ID")
		self.eventData.itemid = lootItemID
		if not(lootItemID) then return end -- bail if bad itemid
		local lootAmount = self:GetItemData(nil, lootText, nil, "QUANTITY")
		self.eventData.qty = lootAmount
		local lootValue = self:GetItemData(lootItemID, nil, lootLink, "VALUE")
		self.eventData.cval = lootValue
		-- Add it to our session totals
		if lootItemID and lootAmount and lootValue then
			self.eventData.calledStore = true
			self:StoreItemValue(lootItemID, lootAmount, lootValue, false)
		end
	end
	self:UpdateText()
end
function NiHao:GetItemLink(lootText)
	local itemLink = nil
	_, _, itemLink = string.find(lootText, "(|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+|h[^|]+|h|r)")
	return itemLink
end
function NiHao:StoreItemValue(itemID, amount, value, reset)
	if reset then
		self.session.itemsEarned[itemID].quantity = 0 -- zeroes out if we're recalculating total value rather than adding
	end
	local lootTotalValue = 0
	if ((amount and value) and amount > 1) then
		-- we have a price and a quantity and that quantity is at least 2
		lootTotalValue = amount * value
	elseif (((not amount) or amount < 2) and value) then
		-- could not determine how many, but we got a vendor price, so consider it just 1 item (missing [item]x# denotes single item
		amount = 1
		lootTotalValue = value
	else
		-- could not figure out anything, bail.	perhaps do something later
		return
	end
	if lootTotalValue > 0 then
		self.session.itemsTotal = self.session.itemsTotal + lootTotalValue
		if not self.session.itemsEarned[itemID] then -- Never seen the item this session, generate new values
			self.session.itemsEarned[itemID] = {
				quantity = 0,
				valueEach = 0,
				valueTotal = 0,
			}
		end
		self.session.itemsEarned[itemID].quantity = self.session.itemsEarned[itemID].quantity + amount
		self.session.itemsEarned[itemID].valueEach = value
		self.session.itemsEarned[itemID].valueTotal = self.session.itemsEarned[itemID].quantity * value
		
		-- Store our indexed itemid table value if not present
		local existsAlready = false
		for k,v in pairs(self.itemsIndexed) do
			if v == itemID then
				existsAlready = true
			end
		end
		if not(existsAlready) then
			table.insert(self.itemsIndexed, itemID)
		end
        -- Sort index by value
        table.sort(self.itemsIndexed, function(s1, s2) return self.session.itemsEarned[s1].valueTotal > self.session.itemsEarned[s2].valueTotal end)
	end
end
function NiHao:RecalculateItems()
	-- Resets the total value, then zeroes out each item in the array and sets their values again
	self.session.itemsTotal = 0
	for k, v in pairs(self.session.itemsEarned) do
		local lootValue = tonumber(self:GetItemData(k, nil, nil, "VALUE"))
		self.session.itemsTotal = self.session.itemsTotal + (lootValue * tonumber(self.session.itemsEarned[k].quantity))
	end
    table.sort(self.itemsIndexed, function(s1, s2) return self.session.itemsEarned[s1].valueTotal > self.session.itemsEarned[s2].valueTotal end)
	self:UpdateText()
end
function NiHao:GetItemData(itemID, lootText, itemLink, dataType)
	-- Pass in nil for itemid, lootText, or link if the data requested does not require it.
	-- datatype should be a string: "VENDOR", "VALUE", "NAME", "RARITY", "ID", "QUANTITY"
	if (dataType == "VENDOR") then
		return tonumber(itemPrice:GetPriceById(itemID))
	end
	if (dataType == "VALUE") then		-- Requires itemid and a PURE itemlink
		-- Take the vendor value from itemPrice library
		local lootValue = tonumber(itemPrice:GetPriceById(itemID))
		--self:Print("Vendor Buyback: "..lootValue) -- TEST
		-- Check that Auctioneer Advanced is installed and that the user has chosen to use Auctioneer values
		--self:Print("Auctioneer check: "..tostring(_G.AucAdvanced)) -- TEST
		if ((_G.AucAdvanced) and (self.db.profile.useAuctioneer)) then
			if (not(itemLink)) then -- If the user didn't add a link, create our own here
				itemLink = self:GenerateLink(itemID)
			end
			local auctioneerValue = tonumber(self:GetAuctioneerData(itemLink))
			local itemRarity = tonumber(self:GetItemData(itemID, nil, nil, "RARITY"))
			-- Only process if Auctioneer price is higher than vendor buyback and ignore grey(poor) quality items
			--self:Print("Item Rarity: "..itemRarity) -- TEST
			if ((auctioneerValue > lootValue) and (itemRarity > 0)) then
				lootValue = auctioneerValue
				--self:Print("Auc price: "..lootValue) -- TEST
			end
		end
		-- Check if user has given us a value to work with, if so, override other values
		if (self.db.profile.itemData[itemID] and self.db.profile.itemData[itemID].value) then
			lootValue = tonumber(self.db.profile.itemData[itemID].value)
			--self:Print("User override: "..lootValue) -- TEST
		end
		if (not(lootValue) or (lootValue < 1)) then
			lootValue = 0
		end
		return tonumber(lootValue)
	end
	if ((dataType == "NAME") or (dataType == "RARITY")) then
		--itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemID or "itemString" or "itemName" or "itemLink") 
		itemName, _, itemRarity, _, _, _, _, _, _, _ = GetItemInfo(itemID)
		if (dataType == "NAME") then			-- Requires itemid
			return itemName
		end
		if (dataType == "RARITY") then		-- Requires itemid
			return itemRarity
		end
	end
	if (dataType == "ID") then			-- Requires link (can be any string containing a link)
		local theItemID = nil
		if (type(itemLink)~="string") then return nil; end;
		theItemID = itemLink:match("|Hitem:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)|h%[(.-)%]|h");
		return tonumber(theItemID);
	end
	if (dataType == "QUANTITY") then	-- Requires lootText
		local quantity = nil
		_, _, quantity = string.find(lootText, "x(%d*).")
		quantity = tonumber(quantity)
		if not(quantity) or (quantity < 2) then
			quantity = 1
		end
		return tonumber(quantity)
	end
end
function NiHao:GetAuctioneerData(itemLink)
	if not(_G.AucAdvanced) then return 0 end -- Bail if AucAdvanced not detected.
	local marketValue, seenCount = AucAdvanced.API.GetMarketValue(itemLink, AucAdvanced.GetFaction())
	if not(marketValue) then marketValue = 0 end -- Ensure marketValue is an integer
	marketValue = math.floor(marketValue) -- round the value off to nearest copper
	--self:Print("Auctioneer value: "..marketValue)
	return marketValue
end
function NiHao:GenerateLink(itemID)
	-- Creates an item link based on the itemID passed in.
	if (not(itemID)) then return nil end
	local itemName		= self:GetItemData(itemID, nil, nil, "NAME")
	local itemRarity	= self:GetItemData(itemID, nil, nil, "RARITY")
	local rarityColor	= ITEM_QUALITY_COLORS[tonumber(itemRarity)].hex
	local itemLink		= rarityColor.."|Hitem:"..itemID..":0:0:0:0:0:0:0|h[" ..itemName.. "]|h|r";
	return itemLink
end
--
--
--				[[ ***** CASH ***** ]]
function NiHao:ConvertCash(copper, cashFormat)
	--[[Description: Converts an integer representing an amount of copper
		into either:
			* A string:	##g##s##c
			* A table:	{ gold = ##, silver = ##, copper = ## }
		* cashAmount = Any integer representing money in copper-only format (100 would equal 1 silver, for example)
		* cashFormat = Defines the type of data to return. Usage: "TABLE" or "TEXT"
		Reference Notes:
			FFD700 = Gold
			C7C7CF = Silver
			EDA55F = Copper--]]
	local cashTable = {}
	if (not(copper)) or (copper < 1) or (tostring(copper) == "-1.#IND") then -- Prevent oddball and negative amounts
		cashTable = { gold=0, silver=0, copper=0 }
		cashText = "|cFFEDA55F"..cashTable.copper.."c|r"
		if (not(cashFormat)) or (cashFormat == "TABLE") then
			return cashTable
		elseif (cashFormat == "TEXT") then
			return cashText
		end
	end
	cashTable.gold = floor(copper / 10000)
	cashTable.silver = mod(floor(copper / 100), 100)
	cashTable.copper = mod(copper, 100)
	if (not cashFormat) or (cashFormat == "TABLE") then
		return cashTable
	elseif cashFormat == "TEXT" then -- Colorize the text for GSC as well as add any leading zeroes applicable.
		if cashTable.gold > 0 then
			cashText = "|cFFFFD700"..cashTable.gold.."g|r".."|cFFC7C7CF"..string.format("%02d", cashTable.silver).."s|r".."|cFFEDA55F"..string.format("%02d", cashTable.copper).."c|r"
		else
			if cashTable.silver > 0 then
				cashText = "|cFFC7C7CF"..cashTable.silver.."s|r".."|cFFEDA55F"..string.format("%02d", cashTable.copper).."c|r"
			else
				cashText = "|cFFEDA55F"..cashTable.copper.."c|r" -- Always show copper, even if it is zero.
			end
		end
		return cashText
	else
		return nil -- Function not called properly
	end
end
function NiHao:UpdateCash()
	local cashCurrent = GetMoney()
	local cashAdded = cashCurrent - (self.session.cashInitial + self.session.cashEarned) -- Never used? Future use?
	if (cashAdded > 0) then
		-- Player gained money.
		-- If these frames are visible or session is paused, we need to ignore any money gained, as it is most likely from selling
		-- NOTE: Money received in a trade window might slip past this.  Trades not tested yet.
		if (MerchantFrame and MerchantFrame:IsVisible()) or (ClassTrainerFrame and ClassTrainerFrame:IsVisible()) or (TradeSkillFrame and TradeSkillFrame:IsVisible()) or (TradeFrame and TradeFrame:IsVisible()) or (MailFrame and MailFrame:IsVisible()) or (AuctionFrame and AuctionFrame:IsVisible()) or (self.session.paused) then
			self.session.cashInitial = self.session.cashInitial + cashAdded
		end
		self.session.cashEarned = cashCurrent - self.session.cashInitial
	else
		-- Player lost money (usually at a vendor).  cashAdded is negated and then subtracted for readability, not functionality.
		self.session.cashInitial = self.session.cashInitial - (cashAdded*-1)
	end
	self:UpdateText()
end
--
--
--				[[ ***** SESSION ***** ]]
function NiHao:LoadSession(notify, name)
	local currentTime = time()
	local savedDuration
	
	-- Check if we are restoring a saved session or a sticky one.
	-- If there is data to load, zero out the current session and load the data.
	if (name) then
		if self.db.profile.savedSessions[name] then
			self.session = {}
			self.session = TableCopy(self.db.profile.savedSessions[name].sessionData)
		else
			self:Print("No such session exists")
			return true -- No saved data to load under the requested session name
		end
	elseif (self.db.profile.stickySession) then
		if self.db.profile.sessionData.startTime then
			-- Copy our saved data over to the current session data
			self.session = {}
			self.session = TableCopy(self.db.profile.sessionData)
		else
			return true -- No sticky data to load
		end
	else
		return true -- If stickySessions are not enabled and we are not loading a saved session, then return confirmation to reset (should be the only way to have gotten here)
	end
	
	-- Figure out the time duration of the saved data
	if (not(self.session.paused)) then
		savedDuration = self.session.timeSaved - self.session.startTime
	else
		savedDuration = self.session.pauseTime - self.session.startTime
		self.session.pauseTime = currentTime
	end
	
	-- Place the start time in the past in the amount of the duration so far
	-- This "makes up for" the time missing between sessions
	self.session.startTime = currentTime - savedDuration
	
	-- Rebuild item index and then ensure values are up to date
	self:RebuildItemIndex()
	self:RecalculateItems()
	
	-- Inform user
	if notify then
		self:Print("Session data loaded")
	end
	
	-- Start paused, if option is set
	if self.db.profile.startPaused then
		self:PauseSession(false)
	else
		self.session.paused = false
	end
	return false -- Data loaded, return "not okay" to reset
end
function NiHao:RebuildItemIndex()
	-- Checks every itemid in the session's list of items looted
	-- against the item index table.  If it is not there yet, it
	-- is added to the index.
	local existsAlready
	for k, v in pairs(self.session.itemsEarned) do
		existsAlready = false
		
		-- Check if the item is already in the index
		for key,val in pairs(self.itemsIndexed) do
			if val == k then
				existsAlready = true
			end
		end
		
		-- If item is not in the index, insert it
		if not(existsAlready) then
			table.insert(self.itemsIndexed, k)
		end
	end
end
function NiHao:ResetSession(notify, stickyCheck)
	-- Just like the function name, this resets all of the data that has
	-- been gathered during the session, zeroing out values and resetting
	-- the timer, clearing the list of items looted, etc.
	-- At the beginning here, we check to see if it is 'okay' to reset things
	-- This is only done if the function was called with stickyCheck set to
	-- true.  The reason we do this is because we don't want to zero out values
	-- if we just loaded sticky session data.  That would defeat the point!
	local okayToReset
	if stickyCheck then
		okayToReset = self:LoadSession(notify) -- See if session data needs to be loaded
	else
		okayToReset = true -- If no sticky check needed, then just reset
	end
	
	if okayToReset then
		-- Reset the session values
		self.session.cashInitial = GetMoney()
		self.session.cashEarned = 0
		self.session.itemsEarned = {}
		self.session.itemsTotal = 0
		self.itemsIndexed = {}
		self.session.startTime = time()
		if self.db.profile.startPaused then
			self:PauseSession(false)
		else
			self.session.paused = false
		end
		self.session.flashCounter = 0
		
		-- Issue the update
		self:UpdateText()
		
		-- Notify user (This check is to prevent double startup spam)
		if notify then
			self:Print("Session Reset")
		end
	end
end
function NiHao:PauseSession(notify)
	-- Pauses the session
	self.session.paused = true
	self.session.pauseTime = time()
	self.session.flashCounter = 0
	
	if (notify) then
		NiHao:Print("Session Paused")
	end
end
function NiHao:ResumeSession()
	-- When resuming the session from being paused, we have to 
	local timeElapsed = time() - NiHao.session.pauseTime
	self.session.startTime = NiHao.session.startTime + timeElapsed
	self.session.pauseTime = nil
	self.session.paused = false
	self.session.flashCounter = 0
	NiHao:Print("Session Resumed")
end
function NiHao:ConvertTime(seconds)
	local hours = 0
	local minutes = 0
	
	if seconds > 3599 then
		hours = math.floor(seconds/3600)
		seconds = seconds - (hours * 3600)
	end
	if seconds > 59 then
		minutes = math.floor(seconds/60)
		seconds = seconds - (minutes * 60)
	end
	
	theTime = hours..":"..string.format("%02d", minutes)..":"..string.format("%02d", seconds)
	return theTime
end
function NiHao:SaveSession(name)
	if (name) then
		-- User is saving session
		self.db.profile.savedSessions[name] = {}
		self.db.profile.savedSessions[name].sessionData = TableCopy(self.session)
		self.db.profile.savedSessions[name].sessionData.timeSaved = time()
		self:Print("Saved session as: "..name)
	else
		if (self.db.profile.stickySession) then -- Save session data
			-- Empty any currently saved data to prevent any ghosting, then
			-- copy in the current session data and a timestamp for when
			-- it was last saved.
			self.db.profile.sessionData = {}
			self.db.profile.sessionData = TableCopy(self.session)
			self.db.profile.sessionData.timeSaved = time()
		end
	end
	self:UpdateSessionMenu()
end
function NiHao:ListSessions()
	self:Print("|cFF00FF00The following sessions were found:|r")
	local sessionPrintCount = 0
	for k,v in pairs(self.db.profile.savedSessions) do
		sessionPrintCount = sessionPrintCount + 1
		self:Print("|cFF00FF00"..sessionPrintCount..".|r "..k)
	end
	if sessionPrintCount == 0 then
		self:Print("|cFFFF6666Saved session database is empty!|r")
	end
end
function NiHao:SavedSessionsExist()
	-- Returns true if there are at least one saved session, else false.
	local sessionsExist = false
	for k,v in pairs(self.db.profile.savedSessions) do
		sessionsExist = true
	end
	return sessionsExist
end
function NiHao:DeleteSession(full, name)
	if full then
		for k,v in pairs(self.db.profile.savedSessions) do
			self.db.profile.savedSessions[k] = nil
		end
		self:Print("|cFF00FF00Deleted ALL saved sessions.|r")
	end
	if not(full) then
		if self.db.profile.savedSessions[name] then
			self.db.profile.savedSessions[name] = nil
			self:Print("|cFF00FF00Deleted session|r "..name.."|cFF00FF00.|r")
		end
	end
	self:UpdateSessionMenu()
end
function NiHao:UpdateSessionMenu()
	local savedSessionMenu = {}
	local theIndex = 2
	savedSessionMenu[1] = {
		type = 'execute',
		name = "Delete All",
		desc = "Click here to delete all saved sessions.",
		order = 1,
		func = function() NiHao:DeleteSession(true, nil) end,
		hidden = function() return not(NiHao:SavedSessionsExist()) end, -- Hide if there are no saved sessions
	}
	for k,v in pairs(self.db.profile.savedSessions) do -- Rebuild session menu
		savedSessionMenu[theIndex] = {
			type = 'group',
			name = k,
			desc = k,
			order = theIndex,
			args = {
				sessionLoad = {
					type = 'execute',
					name = "Load Session",
					desc = "This will load the saved session, destroying any current session data.",
					order = 5,
					func = function() NiHao:LoadSession(true, k) end
				},
				sessionDelete = {
					type = 'execute',
					name = "Delete Session",
					desc = "Deletes this session",
					order = 10,
					func = function() NiHao:DeleteSession(false, k) end
				},
				sessionOverwrite = {
					type = 'execute',
					name = "Overwrite Session",
					desc = "This option will save the current session on top of this session, removing previous data.",
					order = 15,
					func = function() NiHao:SaveSession(k) end
				}
			}
		}
		theIndex = theIndex + 1
	end
	NiHao.OnMenuRequest.args.session.args.savedSessions.args = savedSessionMenu
end
--
--
--				[[ ***** TEXT PROCESSING ***** ]]
function NiHao:ProcessText(haystack)
	-- Check if we are looting an item.
	local needle = self.lootStrings.mobitem
	self.eventData.needle = needle
	if ((needle) and (haystack)) then
		local isItemLoot = nil
		_, _, isItemLoot = string.find(haystack, needle)
		self.eventData.isItemLoot = isItemLoot
		if isItemLoot then
			self.eventData.calledUpdate = true
			self:UpdateItems(haystack)
		end
		return
	end
	-- Found that this message only appears if auto-looting. Manual loots do not show the message in chat window.
	--[[ Check if we are looting cash.
	needle = self.lootStrings.mobcash.prefix
	self.eventData.needle = needle
	if ((needle) and (haystack)) then
		local isCashLoot = nil
		_, _, isCashLoot = string.find(haystack, needle)
		self.eventData.isCashLoot = isCashLoot
		if isCashLoot then
			-- Check cash values
			needle = self.lootStrings.mobcash.copper
			_, _, lootCopper = string.find(haystack, needle)
			needle = self.lootStrings.mobcash.silver
			_, _, lootSilver = string.find(haystack, needle)
			needle = self.lootStrings.mobcash.gold
			_, _, lootGold = string.find(haystack, needle)
			-- Convert nil values to zero
			if not(lootCopper) then lootCopper = 0 end
			if not(lootSilver) then lootSilver = 0 end
			if not(lootGold) then lootGold = 0 end
			amountLooted = lootCopper + (lootSilver * 100) + (lootGold * 10000)
			self.eventData.amountLooted = amountLooted
			self.eventData.calledCashUpdate = true
			self:UpdateCash()
		end
	end--]]
end
function NiHao:ClearSplitItem(split, item)
	if item and split and split[item] then
		local seps = self.Separators[item]
		split[item] = ""
		if seps then
			for k,v in pairs(split) do
				if strlower(k) == seps then
					v = ""
				end
			end
		end
	end
end
function NiHao:SplitChatMessage(frame, event)
	self:ClearChatSections(self.SplitMessage)
	if ( strsub(safestr(event), 1, 8) == "CHAT_MSG" ) then
		local type = strsub(event, 10);
		local info = ChatTypeInfo[type];
		local s = self.SplitMessage
		s.CHATTYPE = type
		s.MESSAGE = safestr(arg1)
		local chatget = getglobal("CHAT_"..type.."_GET")
		if chatget then
			s.TYPEPREFIX, s.TYPEPOSTFIX = string.match(TEXT(chatget), "(.*)%%s(.*)")
		end
		s.TYPEPOSTFIX = safestr(s.TYPEPOSTFIX)
		s.TYPEPREFIX = safestr(s.TYPEPREFIX)
		arg2 = safestr(arg2)
		if strlen(arg2) > 0 then
			if ( strsub(type, 1, 7) == "MONSTER" or type == "RAID_BOSS_EMOTE" ) then
				-- no link
			else
				local plr, svr = strsplit("-", arg2)
				s.pP = "["
				s.lL = "|Hplayer:"
				s.PLAYERLINK = arg2
				s.LL = "|h"
				s.PLAYER = plr
				if svr and strlen(svr) > 0 then
					s.sS = "-"
					s.SERVER = svr
				end
				if arg11 then
					s.PLAYERLINKDATA = ":"..safestr(arg11)
				end
				s.Ll = "|h"
				s.Pp = "]"
			end
		end
		arg6 = safestr(arg6)
		if strlen(arg6) > 0 then
			s.fF = ""
			s.FLAG = TEXT(getglobal("CHAT_FLAG_"..arg6))
			s.Ff = ""
		end
		arg3 = safestr(arg3)
		if ( (strlen(arg3) > 0) and (arg3 ~= "Universal") and (arg3 ~= this.defaultLanguage) ) then
			s.gG = "["
			s.LANGUAGE = arg3
			s.Gg = "] "
		else
			s.LANGUAGE_NOSHOW = arg3
		end
		arg9 = safestr(arg9)
		if strlen(arg9) > 0 then
			local bracket, post_bracket = string.match(s.TYPEPREFIX, "%[(.*)%](.*)")
			bracket = safestr(bracket)
			if strlen(bracket) > 0 then
				s.cC = "["
				s.Cc = "]"
				s.CHANNEL = bracket
				s.TYPEPREFIX = safestr(post_bracket)
			end
			if strlen(safestr(arg8)) > 0 and arg8 > 0 then
				s.CHANNELNUM = tostring(arg8)
				s.CC = ". "
			end
			if arg7 > 0 then
				s.cC = "["
				s.Cc = "] "
				s.CHANNEL, s.zZ, s.ZONE = string.match(arg9, "(.*)(%s%-%s)(.*)")
				if s.CHANNEL == nil then
					s.CHANNEL = arg9
				end
				s.CHANNEL = safestr(s.CHANNEL)
				s.zZ = safestr(s.zZ)
				s.ZONE = safestr(s.ZONE)
				s.Zz = ""
			else
				if strlen(arg9) > 0 then
					s.CHANNEL = arg9
					s.cC = "["
					s.Cc = "] "
				end
			end
		end
		self:ClearChatSections(self.SplitMessageOrg)
		for k,v in pairs(s) do
			self.SplitMessageOrg[k] = v
		end
		if type == "SYSTEM" then
			local pl, p, rest = string.match(s.MESSAGE, "|Hplayer:(.-)|h%[(.-)%]|h(.+)")
			if pl and p then
				local plr, svr = strsplit("-", pl)
				s.pP = "["
				s.lL = "|Hplayer:"
				s.PLAYERLINK = pl
				s.LL = "|h"
				s.PLAYER = plr
				s.Ll = "|h"
				s.Pp = "]"
				s.MESSAGE = rest
				if svr and strlen(svr) > 0 then
					s.sS = "-"
					s.SERVER = svr
				end
				if arg11 then
					s.PLAYERLINKDATA = ":"..safestr(arg11)
				end
			end
		end
		s.ORG = self.SplitMessageOrg
		return self.SplitMessage, info
	end
end
function NiHao:ClearChatSections(message)
	for k,v in pairs(message) do
		message[k] = self.SplitMessageSrc[k] and "" or nil
	end
end
